"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _is = require("./is");

var _is2 = _interopRequireDefault(_is);

var _astLocUtils = require("ast-loc-utils");

var _functions = require("./functions");

var _blocks = require("./blocks");

var _variables = require("./variables");

var _classes = require("./classes");

var _match = require("./match");

var matching = _interopRequireWildcard(_match);

var _comprehension = require("./comprehension");

var comprehension = _interopRequireWildcard(_comprehension);

var _safe = require("./safe");

var _inlinedOperators = require("./inlinedOperators");

var _for = require("./for");

var _helpers = require("./helpers");

var _stdlib = require("./stdlib");

var _config = require("./config");

var _compilerState = require("./compilerState");

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function Lightscript(babel) {
  var t = babel.types;

  // HELPER FUNCTIONS

  function bindMethods(path, methodIds) {
    var assignId = void 0,
        inExpression = false;
    if (path.isClassDeclaration()) {
      assignId = path.node.id;
    } else if (path.parentPath.isAssignmentExpression() && path.parentPath.parentPath.isExpressionStatement()) {
      assignId = path.parentPath.node.left;
    } else if (path.parentPath.isVariableDeclarator()) {
      assignId = path.parentPath.node.id;
    } else {
      var id = path.isClass() ? "class" : "obj";
      assignId = path.getStatementParent().scope.generateDeclaredUidIdentifier(id);
      inExpression = true;
    }
    t.assertOneOf(assignId, ["Identifier", "MemberExpression"]);

    var assignments = methodIds.map(function (methodId) {
      // could be computed, eg `['blah']() => {}`
      t.assertOneOf(methodId, ["Identifier", "Expression"]);
      var isComputed = !t.isIdentifier(methodId);
      var objDotMethod = t.memberExpression(assignId, methodId, isComputed);
      var bind = t.callExpression(t.memberExpression(objDotMethod, t.identifier("bind")), [assignId]);
      return t.assignmentExpression("=", objDotMethod, bind);
    });

    if (inExpression) {
      return path.replaceWith(t.sequenceExpression([t.assignmentExpression("=", assignId, path.node)].concat(_toConsumableArray(assignments), [assignId])));
    } else {
      return path.getStatementParent().insertAfter(assignments.map(function (a) {
        return t.expressionStatement(a);
      }));
    }
  }

  // traverse as top-level item so as to run before other babel plugins
  // (and avoid traversing any of their output)
  function Program(path, state) {
    var opts = state.opts;
    var compilerState = (0, _compilerState.createCompilerState)({
      programPath: path,
      babel,
      opts,
      file: state.file
    });

    if (!compilerState.isLightScript) return;

    (0, _compilerState.initializeCompilerState)();

    // TODO: helpers -> compilerState
    (0, _helpers.resetHelpers)(path);

    path.traverse({

      ForInArrayStatement(path) {
        if (opts.__linter) {
          (0, _for.lintForInArrayStatement)(path);
        } else {
          (0, _for.transformForInArrayStatement)(path);
        }
      }, ForInObjectStatement(path) {
        if (opts.__linter) {
          (0, _for.lintForInObjectStatement)(path);
        } else {
          (0, _for.transformForInObjectStatement)(path);
        }
      }, ArrayComprehension(path) {
        if (opts.enhancedComprehension) {
          comprehension.transformArrayComprehension(path);
        } else {
          // Shim V1 onto V2
          var node = path.node;
          var loop = node.loop;

          if (loop) {
            delete node.loop;
            // TODO: fix babel patch so there's a builder for this...
            node.elements = [(0, _astLocUtils.placeAtNode)({
              type: "LoopComprehension",
              loop
            }, node)];
            path.replaceWith(node);
            comprehension.transformArrayComprehension(path);
          }
        }
      }, ObjectComprehension(path) {
        if (opts.enhancedComprehension) {
          comprehension.transformObjectComprehension(path);
        } else {
          // Shim V1 onto V2
          var node = path.node;
          var loop = node.loop;

          if (loop) {
            delete node.loop;
            // TODO: fix babel patch so there's a builder for this...
            node.properties = [(0, _astLocUtils.placeAtNode)({
              type: "LoopComprehension",
              loop
            }, node)];
            path.replaceWith(node);
            comprehension.transformObjectComprehension(path);
          }
        }
      }, TildeCallExpression: {
        // run on exit instead of enter so that SafeMemberExpression
        // can process differently from a wrapping CallExpression
        // eg; `a?.b~c()` -> `a == null ? null : c(a.b)`
        exit(path) {
          var node = path.node;

          var args = [node.left].concat(_toConsumableArray(node.arguments));
          var callExpr = (0, _astLocUtils.placeAtNode)(t.callExpression(node.right, args), node);

          if (compilerState.inlinedOperatorsEnabled && (0, _inlinedOperators.replaceWithInlinedOperator)(path, path.node.right, args)) return;

          if (path.node.safe) {
            (0, _safe.replaceWithSafeCall)(path, callExpr);
          } else {
            path.replaceWith(callExpr);
          }
        } },

      CallExpression: {
        exit(path) {
          if (compilerState.inlinedOperatorsEnabled && (0, _inlinedOperators.replaceWithInlinedOperator)(path, path.node.callee, path.node.arguments)) return;
          if (path.node.safe) {
            (0, _safe.replaceWithSafeCall)(path, path.node);
          }
        }
      },

      NamedArrowFunction(path) {
        if (path.node.skinny) {
          (0, _functions.replaceWithPlainFunction)(path);
        } else if (path.node.generator) {
          // there are no arrow-generators in ES6, so can't compile to arrow
          (0, _functions.replaceWithBoundFunction)(path);
        } else {
          (0, _functions.replaceWithArrowFunction)(path);
        }
      },

      NamedArrowMemberExpression(path) {
        var object = path.node.object;
        var node = path.node;
        delete node.object;
        node.type = "NamedArrowExpression";

        if (!node.skinny) {
          node.skinny = true; // binding here, don't turn into arrow
          node = (0, _astLocUtils.placeTreeAtLocWhenUnplaced)(t.callExpression(t.memberExpression(node, t.identifier("bind")), [object]), (0, _astLocUtils.getLoc)(object));
        }

        path.replaceWith((0, _astLocUtils.placeAtNode)(t.assignmentExpression("=", (0, _astLocUtils.placeAtLoc)(t.memberExpression(object, path.node.id), (0, _astLocUtils.getSurroundingLoc)([object, path.node.id])), node), node));
      },

      ArrowFunctionExpression(path) {
        if (path.node.skinny) {
          (0, _functions.replaceWithPlainFunction)(path);
        } else if (path.node.generator) {
          (0, _functions.replaceWithBoundFunction)(path);
        }
      },

      Method(path) {
        if ((0, _functions.isNamedArrowFunction)(path.node)) {
          // XXX: non-idiomatic ast mutation, could cause problems...
          path.node.body = (0, _blocks.toBlockStatement)(path.node.body);
        }
      },

      ClassBody(path) {
        var fatArrows = [],
            fatStaticArrows = [],
            constructorPath = void 0;
        path.node.body.forEach(function (method, i) {
          if (!t.isMethod(method)) return;

          if (method.kind === "constructor") {
            return constructorPath = path.get(`body.${i}.body`);
          } else if (method.static && method.skinny === false) {
            fatStaticArrows.push(method.key);
            return method.skinny = true; // prevent infinite recursion
          } else if (method.skinny === false) {
            fatArrows.push(method.key);
            return method.skinny = true; // prevent infinite recursion
          }
        });

        var maybeAddSuper = path.parentPath.node.superClass && constructorPath;
        if (fatArrows.length || maybeAddSuper) {
          constructorPath = (0, _classes.ensureConstructor)(path.parentPath, constructorPath, true);
        }

        if (fatArrows.length) {
          (0, _classes.bindMethodsInConstructor)(path.parentPath, constructorPath, fatArrows);
        }

        if (fatStaticArrows.length) {
          bindMethods(path.parentPath, fatStaticArrows);
        }
      },

      ObjectExpression(path) {
        var fatArrows = [];
        path.node.properties.forEach(function (prop) {
          if (t.isMethod(prop) && prop.skinny === false) {
            fatArrows.push(prop.key);
            // bit ugly, but need a way to ensure we don't double-recurse...
            return prop.skinny = true;
          }
        });

        if (fatArrows.length) {
          bindMethods(path, fatArrows);
        }
      },

      Function: {
        exit(path) {
          (0, _functions.addImplicitReturns)(path);

          // As this is an exit visitor, other LSC transforms have reduced
          // arrows to plain FunctionDeclarations by this point.
          if (path.node.type === "FunctionDeclaration") {
            // somehow this wasn't being done... may signal deeper issues...
            path.getFunctionParent().scope.registerDeclaration(path);
          }
        }
      },

      IfExpression(path) {
        // TODO: better source mapping here -- place implicit null at end of consequent?
        var consequent = (0, _blocks.blockToExpression)(path.get("consequent"));

        var alternate = void 0;
        if (path.node.alternate) {
          alternate = (0, _blocks.blockToExpression)(path.get("alternate"));
        } else {
          alternate = (0, _astLocUtils.placeAtNode)(t.nullLiteral(), path.node);
        }

        path.replaceWith((0, _astLocUtils.placeAtNode)(t.conditionalExpression(path.node.test, consequent, alternate), path.node));
      },

      AssignmentExpression(path) {
        if (path.node.operator === "<-" || path.node.operator === "<!-") {
          path.node.operator = "=";
        }

        // TODO: consider enforcing `now` for MemberExpression too
        if ((0, _is2.default)("MemberExpression", path.node.left)) return;

        if (path.node.isNowAssign === false) {
          throw path.buildCodeFrameError("Incorrect assignment: to reassign, use `now`; to assign as `const`, put on its own line.");
        }
      },

      SafeAwaitExpression(path) {
        var errId = path.scope.generateUidIdentifier("err");
        var awaitIife = (0, _blocks.iife)(t.tryStatement(t.blockStatement([t.returnStatement(t.awaitExpression(path.node.argument))]), t.catchClause(errId, t.blockStatement([t.returnStatement(errId)]))), true);

        // TODO: returntype annotation
        var awaitExpr = (0, _astLocUtils.placeTreeAtLocWhenUnplaced)(t.awaitExpression(awaitIife), (0, _astLocUtils.getLoc)(path.node));
        path.replaceWith(awaitExpr);
      }, SafeMemberExpression(path) {
        (0, _safe.transformSafeMemberExpression)(path);
      }, AwaitExpression(path) {
        if (path.get("argument").isArrayExpression() || path.node.argument.type === "ArrayComprehension") {
          var promiseDotAllCall = t.callExpression(t.memberExpression(t.identifier("Promise"), t.identifier("all")), [path.node.argument]);
          path.get("argument").replaceWith(promiseDotAllCall);
        }
      },

      VariableDeclaration(path) {
        // Error on auto-const when shadowing variable
        if (path.node.kind === "const") {
          if (path.node.extra && path.node.extra.implicit === true) {
            var id = (0, _variables.getShadowingIdentifiers)(path)[0];
            if (id) {
              throw path.buildCodeFrameError(`\`${id.name}\` is shadowed from a higher scope. ` + `If you want to reassign the variable, use \`now ${id.name} = ...\`. ` + "If you want to declare a new shadowed \`const\` variable, " + `you must use \`const ${id.name} = ...\` explicitly.`);
            }
          }
        }
      },

      // collect functions to be imported for the stdlib
      ReferencedIdentifier(path) {
        if (!(0, _variables.checkFalsePositiveReference)(path)) {
          (0, _stdlib.markIdentifier)(compilerState.stdlib, path);
        }
      }, ExistentialExpression(path) {
        (0, _safe.transformExistentialExpression)(path);
      },

      MatchExpression(path) {
        matching.transformMatchExpression(path);
      } // match opts.patternMatching:
      //   | "v3": match_v3.transformMatchExpression(path)
      //   | else: match_v4.transformMatchExpression(path)

      , MatchStatement(path) {
        matching.transformMatchStatement(path);
      } // match opts.patternMatching:
      //   | "v3": match_v3.transformMatchStatement(path)
      //   | else: match_v4.transformMatchStatement(path)
    });

    (0, _compilerState.postprocess)(compilerState);
  }

  var visitor = {
    Program
  };

  var plugin = {
    manipulateOptions(opts, parserOpts, file) {
      var fileTypeInfo = (0, _compilerState.getFileTypeInfo)(file);
      if (!fileTypeInfo.isLightScript) return;

      var pluginOpts = (0, _config.locatePluginOpts)(opts, plugin);

      // Set parser to LightScript mode
      opts.parserOpts = opts.parserOpts || {};
      var nextParserOpts = (0, _config.getParserOpts)(pluginOpts, opts.parserOpts);
      parserOpts.parse = nextParserOpts.parse;
      return parserOpts.plugins = nextParserOpts.plugins;
    },

    visitor
  };

  return plugin;
}

exports.default = Lightscript;
module.exports = exports["default"];